;;########################################################################
;; anovamob.lsp
;; Copyright (c) 1993-96 by Forrest W. Young
;; vista analysis of variance model object for nway balanced anova
;;########################################################################

(require "vista")
(load (strcat *vista-dir-name* "nway"))

;;########################################################################
;; uses nway lispstat model object
;;########################################################################

(defun analysis-of-variance 
  (&key
   (data        current-data)
   (title       "Analysis of Variance")
   (name        nil)
   (dialog      nil)
   (interaction nil))
"ViSta function to perform Analysis of Variance.  
With no arguments the current data are analyzed. Keyword arguments are
:INTERACTION followed by t (calculate interaction) or nil (do not calculate, the default);
:DATA followed by the name of the data to be analyzed (default: current-data);
:TITLE followed by a character string (default: Analysis of Variance);
:DIALOG followed by t (to display parameters dialog box) or nil (default)."
  (if (not (eq current-object data)) (setcd data))
  (if (not name) (strcat "ANV-" (send current-data :name)))
    (send anova-model-object-proto 
          :new  interaction 1 data title name dialog))

(defproto anova-model-object-proto 
  '(nway-model data-object interaction ways classes 
               profile-labels line-labels line-xs line-ys
               f-obs f-crit leverages) 
  () mv-model-object-proto)

(defmeth anova-model-object-proto :isnew (interaction &rest args)
  (send self :data-object current-data)
  (send self :model-abbrev "ANV")
  (send self :interaction interaction)
  (send self :ways (send current-data :ways))
  (send self :classes (send current-data :classes)) 
  (apply #'call-next-method args)
  (send self :leverages (send self :make-leverages))
  )

(defmeth anova-model-object-proto :nway-model (&optional (object-id nil set))
"Message args: (&optional object-id)
Sets or retrieves the object identification information of the underlying nway-model that does the analysis and stores the results."
  (if set (setf (slot-value 'nway-model) object-id))
  (slot-value 'nway-model))

(defmeth anova-model-object-proto :data-object (&optional (object-id nil set))
"Message args: (&optional object-id)
Sets or retrieves the object identification of the analyzed data."
  (if set (setf (slot-value 'data-object) object-id))
  (slot-value 'data-object))

(defmeth anova-model-object-proto :interaction (&optional (logical nil set))
"Method args: (&optional logical)
Sets or returns t or nil indicating whether two-way interaction terms used."
  (if set (setf (slot-value 'interaction) logical))
  (slot-value 'interaction))

(defmeth anova-model-object-proto :ways (&optional (string-list nil set))
"Method args: (&optional string-list)
Sets or returns way names."
  (if set (setf (slot-value 'ways) string-list))
  (slot-value 'ways))

(defmeth anova-model-object-proto :classes (&optional (string-list nil set))
"Method args: (&optional string-list)
Sets or returns class names."
  (if set (setf (slot-value 'classes) string-list))
  (slot-value 'classes))

(defmeth anova-model-object-proto :profile-labels
  (&optional (string-list nil set))
"Method args: (&optional string-list)
Sets or returns profile plot x-axis labels."
  (if set (setf (slot-value 'profile-labels) string-list))
  (slot-value 'profile-labels))

(defmeth anova-model-object-proto :line-labels 
  (&optional (string-list nil set))
"Method args: (&optional string-list)
Sets or returns profile plot line labels."
  (if set (setf (slot-value 'line-labels) string-list))
  (slot-value 'line-labels))

(defmeth anova-model-object-proto :line-xs (&optional (number-list nil set))
"Method args: (&optional number-list)
Sets or returns profile plot line label x locations."
  (if set (setf (slot-value 'line-xs) number-list))
  (slot-value 'line-xs))

(defmeth anova-model-object-proto :line-ys (&optional (number-list nil set))
"Method args: (&optional number-list)
Sets or returns profile plot line label y locations."
  (if set (setf (slot-value 'line-ys) number-list))
  (slot-value 'line-ys))

(defmeth anova-model-object-proto :leverages (&optional (matrix nil set))
"Method args: (&optional matrix)
Sets or returns Sall's predictor source variable leverage values."
  (if set (setf (slot-value 'leverages) matrix))
  (slot-value 'leverages))

(defmeth anova-model-object-proto :f-obs (&optional (list nil set))
  (if set (setf (slot-value 'f-obs) list))
  (slot-value 'f-obs))

(defmeth anova-model-object-proto :f-crit (&optional (list nil set))
  (if set (setf (slot-value 'f-crit) list))
  (slot-value 'f-crit))

(defmeth anova-model-object-proto :options () 
  (let ((choice t))
    (when (send self :dialog)
          (when (or (> (send current-data :nways) 2)
                    (and (= (send current-data :nways) 2)
                         (> (send current-data :nobs) 
                            (send current-data :ncells))))
                (setf choice (choose-item-dialog 
                        "Analysis of Variance:
Include Two-Way Interaction Terms?"
                              '("Main Effects Only (No Interactions)"
                                "Main Effects and Two-Way Interactions")))
                (when choice
                      (if (= 0 choice)
                          (send self :interaction nil)
                          (send self :interaction t))))
          choice)))

(defmeth anova-model-object-proto :analysis ()
  (let* ((interaction (send self :interaction))
         (object (nway-model (send current-data :data) 
                             :class-names (send current-data :classes)
                             :way-names (send current-data :ways)
                             :print nil
                             :interaction interaction)))
    
    (when (not object) 
          (error-message "These data are not balanced.  The analysis of variance method is restricted to balanced data.")
          (top-level))
    (send self :nway-model object)
    object))

(defmeth anova-model-object-proto :report (&key (dialog nil))
  (if (not (eq current-object self)) (setcm self))
  (let ((object (send self :nway-model))
        (w nil))
    (setf w (report-header (send self :title)))
    (display-string (format nil "Analysis of Variance~2%") w)
    (display-string (format nil "Model:               ~a~%" 
                            (send self :name )) w)
    (display-string (format nil "Response  Variable:  ~a~%" 
                            (first (send self :variables)))   w)
    (display-string (format nil "Ways:               ~a~%"
                            (send self :ways)) w)
    (display-string (format nil "Classes:            ~a~%"
                            (send self :classes)) w)
    (send object :display w)))

(defmeth anova-model-object-proto :save-model-template (data-object)
"Args: (data-object)
DATA-OBJECT is the object-identification information of a data object. 
The method contains a template for saving the model-object." 
  `(analysis-of-variance    
    :title      ,(send self :title)
    :name       ,(send self :name) 
    :dialog      nil
    :interaction ,(send current-model :interaction)
    :data (data  ,(send data-object :name)
                 :title      ,(send data-object :title)
                 :variables ',(send self :variables)
                 :types     ',(send self :types)
                 :labels    ',(send self :labels)
                 :ways      ',(send self :ways)
                 :classes   ',(send self :classes)
                 :data      ',(send self :data)))
  )

(defmeth anova-model-object-proto :create-data (&key (dialog nil))
  (if (not (eq current-object self)) (setcm self)) 
  (let* ((creator (send *desktop* :selected-icon))
         (data-obj (send self :data-object))
         (nway (send self :nway-model))
         (response-var (first (send self :variables)))
         (y (send nway :y))
         (x (send nway :x))
         (nx (second (size x)))
         (x-var-names (send nway :predictor-names))
         (fit-values (send nway :fit-values))
         (raw-residuals (send nway :raw-residuals))
         (residuals (send nway :residuals))
         (studentized-residuals (send nway :studentized-residuals))
         (externally-studentized-residuals 
                               (send nway :externally-studentized-residuals))
         (cooks-distances (send nway :cooks-distances))
         (data (bind-columns y x fit-values raw-residuals residuals 
                    studentized-residuals externally-studentized-residuals 
                    cooks-distances))
         )
  (data (concatenate 'string "Results-" (send self :name))
   :created creator
   :creator-object self
   :title (concatenate 'string "ANOVA Results From " 
                       (send self :title))
   :data (combine data)
   :variables (combine response-var x-var-names "Fit Values" 
                       "Raw Residuals" "Residuals" 
                       "Studentized Residuals" "Ext Stud Residuals" 
                       "Cooks Distances")
   :labels (repeat (send data-obj :labels) (send data-obj :cellfreqs)))))

(defmeth anova-model-object-proto :lev (sel)
  (let* ((nway (send self :nway-model))
         (ncols (select (array-dimensions (send nway :x)) 1))
         (ident-mat (identity-matrix ncols))
         (nrows nil)
         (mats (mapcar #'array-dimensions (send nway :design-matrices)))
         (x nil)
         (g nil)
         (i 0)
         (j 0)
         (levmat nil)
         (mat nil)
         )
    (dotimes (i (length mats))
             (setf x (remove nil (combine x 
                     (list (select (select mats i) 1))))))
    (when (not (= sel 0))
          (setf nrows (select x (1- sel)))
          (setf g (if (not (= sel 1))
                      (apply #'+ (select x (iseq (1- sel))))))
          (if (= sel 1) (setf g 0))
          (setf levmat (select ident-mat (iseq g (+ g (1- nrows))) (iseq ncols)))
          (setf mat (send self :sall-lev levmat)))
    (when (= sel 0)
          (setf mat (list (send nway :fit-values) (send nway :y))))
    mat))

(defmeth anova-model-object-proto :sall-lev (L)
"Args: L
L is an indicator matrix for a source in the anova design. L has a column for every term in the design. L has one row for each level of a source, a row has a single 1 indicating the factors position in the design."
  (let* ((nway    (send self :nway-model))
         (x       (send nway :x))
         (xpxiLp  (matmult (inverse (matmult (transpose x) x))(transpose L)))
         (xxpxiLp (matmult x xpxiLp))
         (b       (bind-columns (rest (send nway :coef-estimates))))
         (lambda  (matmult (inverse (matmult L xpxiLp)) L b))
         (vx      (+ (mean (send nway :y)) 
                     (combine (matmult xxpxiLp lambda))))
         (vy      (+ (send nway :residuals) vx))
         )
    (list vx vy)))

(defmeth anova-model-object-proto :make-leverages ()
"Args: none
Returns an nobs x nsources matrix of Sall's Leverage values for the predictor source variables."
  (let* ((levs nil)
         (nway-model (send self :nway-model))
         (n-sources 
          (- (length (send nway-model :source-degrees-of-freedom)) 2))
         (n-obs (send nway-model :nobs))
         )
    (dotimes (i n-sources)
             (setf levs (combine levs (first (send self :lev (1+ i))))))
    (transpose (matrix (list n-sources n-obs) (rest levs)))))

(defmeth anova-model-object-proto :grouped-values (values model sel)
  (let* ((indicator 
          (nth sel (send (send model :data-object) :indicator-matrices)))
         (nclasses (second (size indicator)))
         (members nil)
         (grouped-values nil))
    (dotimes (i nclasses)
             (setf members (select values (which (= 1 (col indicator i)))))
             (setf grouped-values (append grouped-values (list members))))
    grouped-values))

(load (strcat *vista-dir-name* "anovavis"))