;;########################################################################
;; pcamob.lsp
;; principal components ViSta model object
;; Copyright (c) 1991-95 by Forrest W. Young
;;########################################################################

(require "vista")

;;------------------------------------------------------------------------
;;define function for constructing a principal component model object
;;------------------------------------------------------------------------

(defun principal-components 
  (&key 
     (data        current-data)
     (title       "Principal Components")
     (name        nil)
     (dialog      nil)
     (covariances nil))
"ViSta function to perform Principal Components analysis.  
With no arguments, calculate correlations among all active numeric variables in the current data and perform a principal components analysis.
Keyword arguments are:
:COVARIANCES followed by t (analyze covariances) or nil (_analyze correlations, the default).
:DATA followed by the name of the data to be analyzed (default: current-data);
:TITLE followed by a character string (default: Principal Components);
:DIALOG followed by t (to display parameters dialog box) or nil (default)."
  (if (not (eq *current-object* data)) (setcd data))
  (send pca-model-object-proto 
        :new (not covariances) 6 data title name dialog))

;;------------------------------------------------------------------------
;;define principal component model prototype object and its slot-accessors
;;------------------------------------------------------------------------

(defproto pca-model-object-proto '(scores coefs eigenvalues svd corr) 
  () mv-model-object-proto)

(defmeth pca-model-object-proto :isnew (corr &rest args)
;;currently, cannot analyze when nvar>nobs even though it does the analysis
;;right, because the report, visualization, and create data are not right
  (cond
    ((> (send current-data :active-nvar '(numeric)) 
        (send current-data :active-nobs))
     (error-message (format nil "Note: Cannot analyze data ~
                with fewer active observations (~d) ~
                than active numeric variables (~d)." 
                        (send current-data :active-nobs) 
                        (send current-data :active-nvar '(numeric))))
     (send *toolbox* :reset-button 6))
    ((< (send current-data :active-nvar '(numeric)) 2)
     (error-message (format nil "Cannot analyze data ~
                that has less than two active numeric variables."))
     (send *toolbox* :reset-button 6))
    (t
     (send self :model-abbrev "PCA")
     (send self :corr corr)
     (apply #'call-next-method args))))

(defmeth pca-model-object-proto :scores (&optional (values nil set))
  (if set (setf (slot-value 'scores) values))
  (slot-value 'scores))

(defmeth pca-model-object-proto :coefs (&optional (values nil set))
    (if set (setf (slot-value 'coefs) values))
    (slot-value 'coefs))

(defmeth pca-model-object-proto :eigenvalues (&optional (values nil set))
    (if set (setf (slot-value 'eigenvalues) values))
    (slot-value 'eigenvalues))

(defmeth pca-model-object-proto :svd (&optional (structure nil set))
  (if set (setf (slot-value 'svd) structure))
  (slot-value 'svd))

(defmeth pca-model-object-proto :corr (&optional (val nil set))
  (if set (setf (slot-value 'corr) val))
  (slot-value 'corr))

(defmeth pca-model-object-proto :options ()
  (when (send self :dialog)
        (let ((result nil)
              (dialog-value (choose-item-dialog
                "Analysis Options:"
              '("Analyze Covariances"
                "Analyze Correlations") :initial 1))
              )
          (when dialog-value
                (when (= 1 dialog-value) (setf result t))
                (send self :corr result))
          dialog-value)))

(defmeth pca-model-object-proto :analysis ()
  (let* ((left-alpha 1)
         (data (send self :data-matrix))
         (nobs (select (array-dimensions data) 0))
         (prepped-data 
           (if (send self :corr) 
               (/ (normalize (center data) 1) (sqrt (1- nobs)))
               (/ (center data) (sqrt (1- nobs)))))
         (svd (sv-decomp2 prepped-data)) 
         (svd (if (< (sum (col (select svd 2) 0)) 0)
                     (list (* -1 (select svd 0))
                           (select svd 1)
                           (* -1 (select svd 2))
                           (select svd 3))
                  svd))
         (scores (matmult (select  svd 0) 
                          (diagonal (^ (select svd 1) left-alpha))))
         (eigenvalues (^ (select svd 1) 2))
         (coefs (matmult (select  svd 2) 
                         (diagonal (^ (select svd 1)(1- left-alpha)))))
         )
    (send self :svd svd)
    (send self :coefs coefs)
    (send self :scores scores)
    (send self :eigenvalues eigenvalues)
    t))

(defmeth pca-model-object-proto :save-model-template (data-object)
"Args: (data-object)
DATA-OBJECT is the object-identification information of a data object. 
The method contains a template for saving the model-object." 
  `(principal-components    
    :title      ,(send self :title)
    :name       ,(send self :name) 
    :dialog      nil
    :covariances ,(not (send self :corr))
    :data (data  ,(send data-object :name)
                 :title      ,(send data-object :title)
                 :variables ',(send self :variables)
                 :types     ',(send self :types)
                 :labels    ',(send self :labels)
                 :data      ',(send self :data)))
  )

(defmeth pca-model-object-proto :create-data 
  (&key (dialog nil)
        (scores t)
        (coefs  t)
        (input  nil))
  (if (not (eq current-object self)) (setcm self)) 
  (let ((creator (send *desktop* :selected-icon))
        (desires (list (list (if scores 0) (if coefs 1) (if input 2))))
        )
    (if dialog
        (setf desires 
              (choose-subset-dialog "Choose Desired Data Objects"
                  '("Component Scores" 
                    "Component Coefficients"
                    "Analyzed Input Data")
                   :initial (select desires 0))))
    (when desires
          (when (member '0 (select desires 0))
                (send current-model :pca-scores-data-object creator))
          (when (member '1 (select desires 0))
                (send current-model :pca-coefs-data-object  creator))
          (when (member '2 (select desires 0))
                (send current-model :create-input-data-object "PCA" creator)))
    (not (not desires))))

(defmeth pca-model-object-proto :pca-scores-data-object (creator)
  (data (strcat "Scores-" (send self :name))
   :created creator
   :creator-object self
   :title (strcat "PCA Scores for " (send self :title))
   :data (combine (send self :scores))
   :variables (mapcar #'(lambda (x) (format nil "PC~a" x)) 
                      (iseq (min (send self :nvar) (send self :nobs))))
   :labels (send self :labels)
   :types (repeat "Numeric" (min (send self :nvar) (send self :nobs)))
  ))

(defmeth pca-model-object-proto :pca-coefs-data-object (creator)
  (data (concatenate 'string "Coefs-" (send self :name))
   :created creator
   :creator-object self
   :title (concatenate 'string "PCA Coefficients for " 
                       (send self :title))
   :data (combine (send self :coefs))
   :labels (send self :variables)
   :variables (mapcar #'(lambda (x) (format nil "PC~a" x)) 
                      (iseq (send self :nvar)))
   :types (send self :types))
  )

(defmeth pca-model-object-proto :report (&key (dialog nil))
  (if (not (eq current-object self)) (setcm self)) 
  (let* ((w (report-header (send self :title)))
         (labels (send self :labels))
         (vars   (send self :variables))
         (scores (send self :scores))
         (coefs  (transpose (send self :coefs)))
         (eigenvalues (send self :eigenvalues))
         (proportions (/ eigenvalues (sum eigenvalues)))
         (fitmat (transpose 
             (matrix (list 3 (min (send self :nobs) (send self :nvar))) 
                     (combine eigenvalues proportions
                              (cumsum proportions)))))
         (lc-names (mapcar #'(lambda (x) (format nil "PC~a" x)) 
                      (iseq (send current-model :nvar))))) 
    (display-string 
     (format nil "Principal Components Analysis of Variable") w)
    (if (send self :corr) 
        (display-string (format nil " Correlations~2%") w)
        (display-string (format nil " Covariances~2%") w))
    (display-string (format nil "Model: ~a~2%" (send self :name )) w)
    
    (display-string (format nil "Variable Names: ~a~2%" vars) w)
    (display-string (format nil "Fit Indices for each Component:~
       ~%Eigenvalue (amount of total data variance fit by each component)~
       ~%Proportion (of total data variance fit by each component)~
       ~%Cumulative Proportion (of total data variance fit by the components)~
       ~2% Eigenvalue  Proportion  Cum Propor Component~%") w)
    (print-matrix-to-window fitmat w :labels lc-names :decimals 5)
    (display-string (format nil "~%Coefficients (EigenVectors):~%") w)
    (print-matrix-to-window coefs w :labels lc-names :decimals 3)
    (display-string (format nil "~%Component Scores:~%") w)
    (print-matrix-to-window scores w :labels labels :decimals 3) 
    w))    
    
(defmeth pca-model-object-proto :visualize ()
  (if (not (eq current-object self)) (setcm self))
  (let* ((scores (send self :scores))
         (coefs  (send self :coefs))
         (nvar (send self :nvar))
         (ndim (min nvar 9))
         (rays (row-list (select (transpose (send self :coefs) )
                                  (iseq ndim) (iseq nvar))))
         (ratio (sqrt (/ (max (mapcar #'ssq (row-list scores)))
                         (max (mapcar #'ssq (row-list coefs ))))))
         (rays (* rays ratio))
         (labels (send self :variables))
         (model-p t)
         (spinner nil)
         (master nil)
         (scree-location nil)
         (scree-size nil))  
    (spreadplot (column-list scores)
                (send current-model :labels)
                '("PC0 " "PC1" "PC2" "PC3" "PC4" "PC5" "PC6" "PC7" "PC8")
                :scale-type 'fixed 
                :show-histogram nil
                :show-pca-scree t
                :data-object nil)
    (when (> nvar 2)
          (setf spinner spin-plot)
          (setf master  scatmat)
          (setf scree-size graph-size)
          (setf scree-location location22) 
          (send spin-plot :start-buffering)
          (send spin-plot :depth-cuing nil)
          (send spin-plot :scale-type nil)
          (send spin-plot :add-rays rays :ray-labels labels)
          (send spin-plot :scale (iseq ndim) .5)
          (send spin-plot :point-symbol 
                (iseq (send spin-plot :num-points)) 'dot4)
          (send spin-plot :redraw) 
          (send spin-plot :buffer-to-screen)
          (send scatterplot :start-buffering)
          (send scatterplot :title "Biplot") 
          (send scatterplot :add-rays rays :ray-labels labels)
          (send scatterplot :scale (iseq ndim) .5)
          (send scatterplot :point-symbol 
                (iseq (send spin-plot :num-points)) 'dot4)
          (send scatterplot :redraw)
          (send scatterplot :buffer-to-screen)
          (send list-obs :add-points nvar :point-labels labels)
          (let ((proportions (/ (send current-model :eigenvalues) 
                              (send current-model :nvar))))
            (setf pca-scree-plot
                 (scree-plot proportions
                             :location scree-location
                             :size scree-size))
            (send pca-scree-plot :variable-label 1 "Proportion")
            (send pca-scree-plot :range 1 0 
                (/ (ceiling (* 10 (select proportions 0))) 10) )
            (send pca-scree-plot :show-window)
            (send pca-scree-plot :menu nil)
            (defmeth pca-scree-plot :close ()
              (send scatmat :close))))))